let map;
let markers = []; // Array para armazenar os objetos dos marcadores
let markersData = {}; // Armazenar as informaes de lat/lng e tipo para comparar mudanas futuras

// Coordenadas fixas que sero usadas
const predefinedCoordinates = {
    buraco: { lat: -3.14, lng: -60.02 },
    acidente: { lat: -3.26, lng: -60.50 },
    trafego: { lat: -3.32, lng: -60.75 },
    profundidade: { lat: -3.45, lng: -60.90 }
};

// Funo para inicializar o mapa
window.initMap = function() {
    map = new google.maps.Map(document.getElementById("map"), {
        zoom: 7,
        center: { lat: -3.14, lng: -60.02 },
        mapTypeId: 'hybrid' // Exibe mapa hbrido (satlite + rtulos)
    });

    // Carregar os marcadores do banco de dados via API REST
    loadMarkersFromAPI();

    // Iniciar polling para atualizar marcadores a cada 10 segundos
    setInterval(loadMarkersFromAPI, 10000);
};

// Funo para carregar os marcadores via API REST
function loadMarkersFromAPI() {
    fetch(wpApiSettings.root + 'wp/v2/marcadores')
        .then(response => response.json())
        .then(data => {
            if (Array.isArray(data)) {
                data.forEach(marcador => {
                    const lat = parseFloat(marcador.meta.lat);
                    const lng = parseFloat(marcador.meta.lng);
                    const tipo = marcador.meta.tipo;
                    const profundidade = marcador.meta.profundidade || 'N/A';
                    const identificacao = marcador.meta.identificacao || 'Desconhecido';
                    const id = marcador.id;

                    // Verificar se o marcador j existe e se mudou
                    if (!markersData[id] || markersData[id].lat !== lat || markersData[id].lng !== lng) {
                        adicionarMarcador(id, lat, lng, tipo, profundidade, identificacao);
                        markersData[id] = { lat, lng, tipo }; // Atualizar o armazenamento local
                    }
                });

                // Remover os marcadores que foram deletados no servidor
                markers.forEach((markerObj, index) => {
                    if (!data.find(m => m.id === markerObj.id)) {
                        markerObj.marker.setMap(null); // Remove o marcador do mapa
                        markers.splice(index, 1); // Remove do array de marcadores
                        delete markersData[markerObj.id]; // Remove do armazenamento local
                    }
                });
            }
        })
        .catch(error => {
            console.error("Erro ao carregar marcadores:", error);
        });
}

// Funo para adicionar um marcador com coordenadas fixas
function addMarkerWithFixedCoordinates(type) {
    const coordinates = predefinedCoordinates[type];

    // Enviar dados para a API REST para salvar o marcador
    fetch(wpApiSettings.root + 'wp/v2/marcadores', {
        method: 'POST',
        headers: {
            'Content-Type': 'application/json',
            'X-WP-Nonce': wpApiSettings.nonce
        },
        body: JSON.stringify({
            title: `${type} - ${wpApiSettings.identificacao}`,
            meta: {
                lat: coordinates.lat,
                lng: coordinates.lng,
                tipo: type,
                identificacao: wpApiSettings.identificacao, // Inclui a identificao do usurio
                profundidade: 'N/A'
            },
            status: 'publish'
        })
    })
    .then(response => response.json())
    .then(data => {
        const id = data.id;
        if (!id || id === 0) {
            throw new Error('Erro: ID no foi retornado corretamente.');
        }

        adicionarMarcador(id, coordinates.lat, coordinates.lng, type, 'N/A', wpApiSettings.identificacao);
    })
    .catch(error => {
        console.error("Erro ao salvar marcador:", error);
    });
}

// Funo para adicionar o marcador ao mapa e exibir infowindow
function adicionarMarcador(id, lat, lng, tipo, profundidade = 'N/A', identificacao = 'Desconhecido') {
    const marker = new google.maps.Marker({
        position: { lat: lat, lng: lng },
        map: map,
        icon: {
            url: getIconUrl(tipo),
            scaledSize: new google.maps.Size(40, 40)
        }
    });

    markers.push({ id: id, marker: marker });

    const infowindowContent = `
        <div class="custom-infowindow">
          <h4>Identificao: ${identificacao}</h4>
          <h4>${tipo.charAt(0).toUpperCase() + tipo.slice(1)}</h4>
          <p><strong>Lat/Lng:</strong> ${lat}, ${lng}</p>
          <p><strong>Profundidade:</strong> ${profundidade}</p>
          <button onclick="removerMarcador(${id})">Remover</button>
        </div>`;

    const infowindow = new google.maps.InfoWindow({
        content: infowindowContent
    });

    infowindow.open(map, marker);
}

// Funo para remover um marcador
function removerMarcador(markerId) {
    fetch(`${wpApiSettings.root}wp/v2/marcadores/${markerId}`, {
        method: 'DELETE',
        headers: {
            'Content-Type': 'application/json',
            'X-WP-Nonce': wpApiSettings.nonce
        }
    })
    .then(response => {
        if (!response.ok) {
            throw new Error('Erro ao remover marcador');
        }
        return response.json();
    })
    .then(data => {
        console.log('Marcador removido com sucesso:', data);
        // Remove o marcador do mapa e da lista de marcadores
        const markerIndex = markers.findIndex(m => m.id === markerId);
        if (markerIndex !== -1) {
            markers[markerIndex].marker.setMap(null); // Remove do mapa
            markers.splice(markerIndex, 1); // Remove da lista
        }
    })
    .catch(error => {
        console.error('Erro ao remover marcador:', error);
    });
}

// Funo para obter a URL do cone com base no tipo de marcador
function getIconUrl(type) {
    switch(type) {
        case 'buraco':
            return 'https://maps.google.com/mapfiles/ms/icons/orange-dot.png';
        case 'acidente':
            return 'https://maps.google.com/mapfiles/ms/icons/blue-dot.png';
        case 'trafego':
            return 'https://maps.google.com/mapfiles/ms/icons/green-dot.png';
        case 'profundidade':
            return 'https://maps.google.com/mapfiles/ms/icons/purple-dot.png';
        default:
            return 'https://maps.google.com/mapfiles/ms/icons/red-dot.png';
    }
}
